/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * Use of this source code is governed by a BSD-style license that can be
 * found in the license.txt file.
 */

/** @file
 * @brief RTC/PPI/ADC modules example
 *
 * This example enables the RTC with a TICK frequency of 10Hz. Each RTC TICK triggers
 * ADC conversion through a PPI channel. The supply voltage for the chip is sampled
 * with the ADC. The supply voltage provided by USB is approximately 3.0 volts, so the
 * observed value should be around 212 with the configured 8-bit ADC resolution.
 */

#include <stdbool.h>
#include "nrf.h"
#include "nrf_gpio.h"
#include "SEGGER_RTT.h"
#include "boards.h"

#define GPIO_TOGGLE_TICK_EVENT    (8)                       /*!< Pin number to toggle when there is a tick event in RTC */
#define ADC_RESOLUTION            ADC_CONFIG_RES_10bit

/** Configures port 1 as output
 */
static void gpio_config(void)
{
	// Configure LED-pins as outputs.
	LEDS_CONFIGURE(LEDS_MASK);
}

/** Configures and starts the low frequency 32kHz clock (required to operate the RTC)
 */
static void lfclk_config(void)
{
  NRF_CLOCK->LFCLKSRC = (CLOCK_LFCLKSRC_SRC_Xtal << CLOCK_LFCLKSRC_SRC_Pos);
  NRF_CLOCK->EVENTS_LFCLKSTARTED = 0;
  NRF_CLOCK->TASKS_LFCLKSTART = 1;
	
	// Wait for the low frequency clock to start
  while (NRF_CLOCK->EVENTS_LFCLKSTARTED == 0) {}
  NRF_CLOCK->EVENTS_LFCLKSTARTED = 0;
}

/** Configures and starts RTC
 */
static void rtc_config(void)
{
  NRF_RTC0->PRESCALER = 4095;                   // Set prescaler
	NRF_RTC0->EVTENSET = RTC_EVTENSET_COMPARE0_Msk;
	NRF_RTC0->CC[0] = 4;
	NRF_RTC0->TASKS_START = 1;                                 // Start RTC0
}

/** Configures and enables a PPI channel
 */
static void ppi_init(void)
{
  // Configure PPI channel 0 to start ADC task on RTC0 compare 0 event
	NRF_PPI->CH[0].EEP = (uint32_t)&NRF_RTC0->EVENTS_COMPARE[0];
  NRF_PPI->CH[0].TEP = (uint32_t)&NRF_ADC->TASKS_START;
	
	// Configure PPI channel 1 to clear the RTC0 timer on RTC0 compare 0 event
	NRF_PPI->CH[1].EEP = (uint32_t)&NRF_RTC0->EVENTS_COMPARE[0];
  NRF_PPI->CH[1].TEP = (uint32_t)&NRF_RTC0->TASKS_CLEAR;

  // Enable configured PPI channels
  NRF_PPI->CHEN = (PPI_CHEN_CH0_Enabled << PPI_CHEN_CH0_Pos);
	NRF_PPI->CHEN |= (PPI_CHEN_CH1_Enabled << PPI_CHEN_CH1_Pos);
}

/** Configures and enables the ADC
 */
void ADC_init(void)
{	
	/* Enable interrupt on ADC sample ready event*/		
	NRF_ADC->INTENSET = ADC_INTENSET_END_Msk;
	NVIC_EnableIRQ(ADC_IRQn);	
	
	/* Configure ADC - set reference input source to internal 1.2V bandgap */
	NRF_ADC->CONFIG = (ADC_CONFIG_REFSEL_VBG << ADC_CONFIG_REFSEL_Pos);
	/* Configure ADC - set input source to VDD/3 */
	NRF_ADC->CONFIG |= (ADC_CONFIG_INPSEL_SupplyOneThirdPrescaling << ADC_CONFIG_INPSEL_Pos);
	/* Configure ADC - select 8 bit resolution */
	NRF_ADC->CONFIG |= (ADC_RESOLUTION << ADC_CONFIG_RES_Pos);
	/* Enable ADC*/
	NRF_ADC->ENABLE = ADC_ENABLE_ENABLE_Enabled;	
}

void write_adc_result_to_leds(uint16_t adc_result)
{
  // Write 3 LSB ADC result to the LEDs on nRF51-Dongle (PCA10031) and nRF51-DK (PCA10028)
	nrf_gpio_pin_write(LED_1, (adc_result & 0x01));
	nrf_gpio_pin_write(LED_2, (adc_result & 0x02));
	nrf_gpio_pin_write(LED_3, (adc_result & 0x04));

	//If this is the nRF51DK (PCA10028) with 4 leds, write LED_4 also
	if (LEDS_NUMBER == 4) 
	{ 
		nrf_gpio_pin_write(LED_4, (NRF_ADC->RESULT & 0x08));
	}
}

bool adc_calibrate(uint16_t adc_result, uint16_t * adc_result_calibrated, uint8_t adc_resolution)
{
	uint32_t uicr_value;
	uint8_t offset_error;
	uint8_t gain_error;
	bool is_calibrated;
	
	// Apply ADC calibration described on 
	// https://devzone.nordicsemi.com/question/21653/how-to-calibrate-the-nrf51-adc-to-correct-offset-and-gain-error/
	uicr_value = *(uint32_t *)0x10000024;       //Read ADC gain and offset error from UICR
	offset_error = uicr_value;
	gain_error = uicr_value >> 8;
	
	if(adc_resolution == ADC_CONFIG_RES_10bit)
	{
		*adc_result_calibrated = adc_result * (1024 + gain_error) / 1024 + offset_error;    //calibrate
		is_calibrated = true;
	}
	else
	{
		//Calibration method not valid for given ADC resolution. Return adc value uncalibrated.
		*adc_result_calibrated = adc_result;      
		is_calibrated = false;
	}
	return is_calibrated;
}	

/* Interrupt handler for ADC data ready event */
void ADC_IRQHandler(void)
{
	uint16_t adc_result_calibrated;

	/* Clear dataready event */
  NRF_ADC->EVENTS_END = 0;	
	
	//Attempt to calibrate the ADC result
	adc_calibrate(NRF_ADC->RESULT, &adc_result_calibrated, ADC_RESOLUTION);

	//Print ADC result on LEDs
	write_adc_result_to_leds(adc_result_calibrated);

	//print ADC result on RTT, open J-Link RTT Viewer on port 0 to see the ADC result.
	SEGGER_RTT_printf(0, "ADC result:  %d\n", adc_result_calibrated);	
}


/**
 * main function
 */
int main(void)
{
  gpio_config();
	lfclk_config();
  rtc_config();
	ppi_init();
	ADC_init();
	
	SEGGER_RTT_WriteString(0, "Device starting ...\n");
  while (true)
  {	
		//Put CPU to sleep while waiting for interrupt
		__WFE();
		__SEV();
		__WFE();
  }
}

/**
 *@}
 **/
